from __future__ import annotations

import typing

import flood
from flood import spec

if typing.TYPE_CHECKING:
    import types

    import toolcli


styles: toolcli.StyleTheme = {
    'title': 'bold #00e100',
    'metavar': 'bold #e5e9f0',
    'description': '#aaaaaa',
    'content': '#00B400',
    'option': 'bold #e5e9f0',
    'comment': '#888888',
}


plot_colors = {
    'green_shades': [
        'forestgreen',
        'limegreen',
        'chartreuse',
    ],
    'red_shades': [
        'firebrick',
        'red',
        'salmon',
    ],
    'purple_shades': [
        'rebeccapurple',
        'blueviolet',
        'mediumslateblue',
    ],
    'orange_shades': [
        'darkgoldenrod',
        'darkorange',
        'gold',
    ],
    'blue_shades': [
        'blue',
        'dodgerblue',
        'lightskyblue',
    ],
    'streetlight': [
        'crimson',
        'goldenrod',
        'limegreen',
    ],
}


def get_nodes_plot_colors(
    nodes: typing.Mapping[str, flood.Node]
) -> typing.Mapping[str, str]:
    colors = {}
    taken = set()
    for node in nodes.values():
        # print version
        version = node['client_version']
        if version is None:
            version = ''

        # decide color
        if (
            (version is not None and 'reth' in version)
            or ('reth' in node['name'])
        ) and 'orange_shades' not in taken:
            color = 'orange_shades'
        elif (
            (version is not None and 'erigon' in version)
            or ('erigon' in node['name'])
        ) and 'blue_shades' not in taken:
            color = 'blue_shades'
        else:
            for color_name in plot_colors.keys():
                if color_name not in taken:
                    color = color_name
                    break
            else:
                raise Exception('out of colors')

        colors[node['name']] = color
        taken.add(color)

    return colors


def _get_tqdm() -> types.ModuleType:
    import sys

    if 'jupyter_client' in sys.modules:
        try:
            import ipywidgets  # type: ignore # noqa: F401
            import tqdm.notebook as tqdm

            return tqdm
        except ImportError:
            pass

    import tqdm  # type: ignore

    return tqdm


def print_metric_tables(
    results: typing.Mapping[str, spec.LoadTestOutput],
    metrics: typing.Sequence[str],
    *,
    decimals: int | None = None,
    comparison: bool = False,
    indent: int | str | None = None,
) -> None:
    import toolstr

    if len(results) == 0:
        toolstr.print('no results', indent=indent)
        print()

    names = list(results.keys())
    rates = results[names[0]]['target_rate']
    for metric in metrics:
        # create labels
        if metric == 'success':
            suffix = ''
        else:
            suffix = ' (s)'
        unitted_names = [name + suffix for name in names]
        labels = ['rate (rps)'] + unitted_names
        if comparison:
            if len(results) != 2:
                raise NotImplementedError('comparison of >2 tests')
            comparison_label = names[0] + ' / ' + names[1]
            labels.append(comparison_label)
        else:
            comparison_label = None

        # build rows
        rows: list[list[typing.Any]] = [[rate] for rate in rates]
        values = []
        for name, result in results.items():
            for row, value in zip(rows, result[metric]):  # type: ignore
                row.append(value)
                values.append(value)
        if comparison:
            for row in rows:
                row.append(row[-2] / row[-1])

        # compute column formats
        if all(value > 1 for value in values):
            use_decimals = 1
        else:
            if decimals is None:
                use_decimals = 6
            else:
                use_decimals = decimals
        column_formats = {
            column: {'decimals': use_decimals} for column in unitted_names
        }
        if comparison_label is not None:
            column_formats[comparison_label] = {
                'decimals': 1,
                'percentage': True,
            }

        # print header
        toolstr.print_text_box(
            toolstr.add_style(metric + ' vs load', flood.styles.get('metavar')),
            style=flood.styles.get('content'),
            indent=indent,
        )

        # print table
        toolstr.print_table(
            rows,
            labels=labels,
            column_formats=column_formats,  # type: ignore
            label_style=flood.styles.get('metavar'),
            border=flood.styles.get('content'),
            indent=indent,
        )
        if metric != metrics[-1]:
            print()


#
# # generic restylings of toolstr functions
#


def print_text_box(text: str) -> None:
    import toolstr

    toolstr.print_text_box(
        text,
        text_style=flood.styles.get('metavar'),
        style=flood.styles.get('content'),
    )


def print_header(text: str) -> None:
    import toolstr

    toolstr.print_header(
        text,
        text_style=flood.styles.get('metavar'),
        style=flood.styles.get('content'),
    )


def print_bullet(*args: typing.Any, **kwargs: typing.Any) -> None:
    import toolstr

    toolstr.print_bullet(
        *args,
        **kwargs,
        styles=flood.styles,
    )


def print_table(*args: typing.Any, **kwargs: typing.Any) -> None:
    import toolstr

    toolstr.print_table(
        *args,
        **kwargs,
        label_style=flood.styles.get('metavar'),
        border=flood.styles.get('content'),
    )


def print_multiline_table(*args: typing.Any, **kwargs: typing.Any) -> None:
    import toolstr

    toolstr.print_multiline_table(
        *args,
        **kwargs,
        label_style=flood.styles.get('metavar'),
        border=flood.styles.get('content'),
    )



from __future__ import annotations

import typing

from flood import spec

if typing.TYPE_CHECKING:
    import numpy as np


def get_rng(random_seed: spec.RandomSeed | None = None) -> np.random.Generator:
    import numpy as np

    if random_seed is None:
        random_seed = 0
    if isinstance(random_seed, int):
        gen = np.random.Generator(np.random.PCG64(random_seed))
    if isinstance(gen, np.random.Generator):
        return gen
    else:
        raise Exception('invalid seed format: ' + str(type(random_seed)))



from __future__ import annotations

import typing

import toolcli

import flood


help_message = """create report of previous test results

outputs a notebook file and an html file in output directory"""


def get_command_spec() -> toolcli.CommandSpec:
    return {
        'f': report_command,
        'help': help_message,
        'args': [
            {
                'name': 'test_dirs',
                'nargs': '+',
                'help': 'space-seperated paths to previous tests',
            },
            {
                'name': ['-o', '--output'],
                'help': 'output directory \[default = current directory]',
            },
            {
                'name': ['-m', '--metrics'],
                'nargs': '+',
                'help': 'metrics to graph in report',
            },
        ],
        'examples': ['tests/test1 tests/test2 tests/test3'],
    }


def report_command(
    test_dirs: typing.Sequence[str],
    output: str | None,
    metrics: typing.Sequence[str] | None,
) -> None:
    if output is None:
        import os

        output = os.getcwd()

    flood.create_load_test_report(
        test_paths=test_dirs,
        output_dir=output,
        metrics=metrics,
    )



"""type definitions for typing annotations"""

from __future__ import annotations

import typing


if typing.TYPE_CHECKING:
    #
    # # generic types
    #

    Call = typing.Mapping[str, typing.Mapping[str, typing.Any]]

    MethodCalls = typing.Mapping[str, typing.Sequence[typing.Any]]

    class Node(typing.TypedDict):
        name: str
        url: str
        remote: str | None
        client_version: str | None
        network: str | int | None

    NodeShorthand = typing.Union[str, Node]

    NodesShorthand = typing.Union[
        typing.Sequence[str],
        typing.Sequence[Node],
        typing.Mapping[str, str],
        typing.Mapping[str, Node],
    ]
    Nodes = typing.Mapping[str, Node]

    class ProgressBar(typing.TypedDict):
        desc: str
        position: int | None
        leave: bool
        colour: str
        disable: bool

    #
    # # generators
    #

    import numpy as np

    RandomSeed = typing.Union[int, np.random._generator.Generator]

    #
    # # latency test types
    #

    NodeMethodLatencies = typing.Mapping[str, typing.Sequence[float]]
    NodesMethodLatencies = typing.Mapping[str, NodeMethodLatencies]

    class LatencyBenchmarkResults(typing.TypedDict):
        nodes: typing.Mapping[str, Node]
        methods: typing.Sequence[str]
        samples: int | None
        calls: typing.Mapping[str, typing.Sequence[str]]
        latencies: NodesMethodLatencies

    #
    # # equality test
    #

    EqualityTest = tuple[
        str,
        typing.Callable[..., typing.Any],
        typing.Sequence[typing.Any],
        typing.Mapping[str, typing.Any],
    ]

    #
    # # load test types
    #

    class VegetaAttack(typing.TypedDict):
        rate: int
        duration: int
        calls: typing.Sequence[typing.Any]
        vegeta_kwargs: typing.Mapping[str, typing.Any]

    VegetaKwargs = typing.Mapping[str, typing.Union[str, None]]
    MultiVegetaKwargs = typing.Sequence[VegetaKwargs]
    VegetaKwargsShorthand = typing.Union[VegetaKwargs, MultiVegetaKwargs]

    LoadTest = typing.Sequence[VegetaAttack]

    class LoadTestColumnWise(typing.TypedDict):
        rates: typing.Sequence[int]
        durations: typing.Sequence[int]
        calls: typing.Sequence[typing.Sequence[typing.Any]]
        vegeta_kwargs: typing.Sequence[typing.Any]

    LoadTestMode = typing.Literal['stress', 'spike', 'soak']

    LoadTestGenerator = typing.Callable[..., LoadTest]
    MultiLoadTestGenerator = typing.Callable[..., typing.Mapping[str, LoadTest]]

    #
    # # load tests outputs
    #

    class RawLoadTestOutputDatum(typing.TypedDict):
        latencies: typing.Mapping[str, float]
        bytes_in: typing.Mapping[str, int]
        bytes_out: typing.Mapping[str, int]
        earliest: str
        latest: str
        end: str
        duration: int
        wait: int
        requests: int
        rate: float
        throughput: float
        success: int
        status_codes: typing.Mapping[str, int]
        errors: typing.Sequence[str]
        first_request_timestamp: str
        last_request_timestamp: str
        last_response_timestamp: str
        final_wait_time: int

    class LoadTestOutputDatum(typing.TypedDict):
        target_rate: int
        actual_rate: float
        target_duration: int
        actual_duration: float
        requests: int
        throughput: float
        success: float
        min: float | None
        mean: float
        p50: float
        p90: float
        p95: float
        p99: float
        max: float
        status_codes: typing.Mapping[str, int]
        errors: typing.Sequence[str]
        first_request_timestamp: str
        last_request_timestamp: str
        last_response_timestamp: str
        final_wait_time: float
        raw_output: str | None

    class LoadTestOutput(typing.TypedDict):
        target_rate: typing.Sequence[int]
        actual_rate: typing.Sequence[float]
        target_duration: typing.Sequence[int]
        actual_duration: typing.Sequence[float]
        requests: typing.Sequence[int]
        throughput: typing.Sequence[float]
        success: typing.Sequence[float]
        min: typing.Sequence[float | None]
        mean: typing.Sequence[float]
        p50: typing.Sequence[float]
        p90: typing.Sequence[float]
        p95: typing.Sequence[float]
        p99: typing.Sequence[float]
        max: typing.Sequence[float]
        status_codes: typing.Sequence[typing.Mapping[str, int]]
        errors: typing.Sequence[typing.Sequence[str]]
        first_request_timestamp: typing.Sequence[str]
        last_request_timestamp: typing.Sequence[str]
        last_response_timestamp: typing.Sequence[str]
        final_wait_time: typing.Sequence[float]
        raw_output: typing.Sequence[str | None]

    RunType = typing.Literal['single_test']

    class SingleRunTestPayload(typing.TypedDict):
        version: str
        type: RunType
        name: str
        test: LoadTest

    class SingleRunResultsPayload(typing.TypedDict):
        version: str
        type: RunType
        nodes: Nodes
        results: typing.Mapping[str, LoadTestOutput]
